/****************************************************************************

 Module
  evnthndl.c
  
 Description
  Event processing routines in oAESis.
  
 Author(s)
 	cg (Christer Gustavsson <d2cg@dtek.chalmers.se>)

 Revision history
 
  960101 cg
   Added standard header.
   Mouse_gain_mctrl() and Mouse_release_mctrl() added.

  960129
   Changed name form mouse.c to evnthndl.c.
   Public routines changed prefix from Mouse_ to Evhd_.
   
 Copyright notice
  The copyright to the program code herein belongs to the authors. It may
  be freely duplicated and distributed without fee, but not charged for.
 
 ****************************************************************************/

/****************************************************************************
 * Used interfaces                                                          *
 ****************************************************************************/

#include <basepage.h>
#include <errno.h>
#include <fcntl.h>
#include <ioctl.h>
#include <mintbind.h>
#include <osbind.h>
#include <signal.h>
#include <stdio.h>
#include <unistd.h>

#include "appl.h"
#include "evnthndl.h"
#include "global.h"
#include "memdefs.h"
#include "mesagdef.h"
#include "misc.h"
#include "mousedev.h"
#include "objc.h"
#include "resource.h"
#include "types.h"
#include "vdi.h"
#include "wind.h"

#include <sysvars.h>

/****************************************************************************
 * Macros                                                                   *
 ****************************************************************************/

#define MCTRL_SEM 0x4d435482L /*"MCTR"*/
#define	MOUSESEM	0x4d4f4556L	/*MOEV*/

#define EVHD_APID  (-2)

/****************************************************************************
 * Module global variables                                                  *
 ****************************************************************************/

static struct {
	WORD       evid;
	WORD       mouseid;
	REVENTLIST *eventlist;

	WORD       lastervalid;
	EVNTREC    last_evntrec;
	WORD       mousefd;
	ULONG      lasttime;
}evntglbl = {
	-1,
	-1,
	0L,
	0,
	{-1,-1},
	-1,
	0UL
};

static CSEMA mctrl = {
	MCTRL_SEM,
	-1L,
	0,
	-1
};

/****************************************************************************
 * Local functions (use static!)                                            *
 ****************************************************************************/


static WORD time_message(EVNTREC *er) {
	ULONG	newtime;
	
  newtime = globals.time;

	if(newtime != evntglbl.lasttime) {
		er->ap_event = APPEVNT_TIMER;
		er->ap_value = newtime - evntglbl.lasttime;
					
		evntglbl.lasttime = newtime;
		
		return 1;
	};
	
	return 0;
}

static LONG	get_evntpacket(EVNTREC *er,WORD maxwait) {
	LONG kb_shift;

	LONG stdinrfds,mrfds;
	LONG rhndl;
  

  if(evntglbl.lastervalid) {
		*er = evntglbl.last_evntrec;
		evntglbl.lastervalid = 0;
		return 1;
	};
	
	stdinrfds = 1L << (LONG)STDIN_FILENO;
	mrfds = 1L << (LONG)evntglbl.mousefd;
	rhndl = stdinrfds | mrfds;

	if(Fselect(maxwait,&rhndl,NULL,0L) == 0) {
		return 0;
	};
		
	if(rhndl & stdinrfds) {
		ULONG c = Fgetchar(STDIN_FILENO,0);
		
		kb_shift = Kbshift(-1);
		
		evntglbl.last_evntrec.ap_event = APPEVNT_KEYBOARD;
		evntglbl.last_evntrec.ap_value = ((c & 0xff) << 16) +
			((c & 0x00ff0000L) << 8) + kb_shift;

	};

	if(rhndl & mrfds)  {
 		Fread(evntglbl.mousefd, sizeof(EVNTREC), 
 		       &evntglbl.last_evntrec);
 	};
	
	if(time_message(er)) {
		evntglbl.lastervalid = 1;
	}
	else {
		*er = evntglbl.last_evntrec;
	};	
	
	return 1;
}

static void	update_mousevalues(EVNTREC *er) {
  switch((WORD)er->ap_event) {
  case	APPEVNT_BUTTON	:
    globals.mouse_button = ((WORD *)&er->ap_value)[1];
    break;	
    
  case	APPEVNT_MOUSE	:
    globals.mouse_x = ((WORD *)&er->ap_value)[1];
    globals.mouse_y = ((WORD *)&er->ap_value)[0];
    break;
  };
}

static void handle_arrow_click(WORD win_id,WORD elem,WORD msg) {
	COMMSG    mesag;
	EVNTREC   er;
	WINSTRUCT *win;
	WORD      owner;
	
	Evhd_gain_mctrl(EVHD_APID,MCTRL_BUTTON);

	mesag.type = WM_ARROWED;
	mesag.sid = 0;
	mesag.length = 0;
	mesag.msg0 = win_id;
	mesag.msg1 = msg;

	Rdwrs_operation(WSTARTWRITE);

	win = findwin(win_id);

	if(elem) {
		win->tree[elem].ob_state |= SELECTED;
		drawwinelem(evntglbl.evid,win,&globals.screen,elem);
	};

	owner = win->owner;

	Rdwrs_operation(WENDWRITE);

	do {
	  if(get_evntpacket(&er,globals.arrowrepeat) == 0) {
			Appl_do_write(owner,16,&mesag);
	  };
	  
	  update_mousevalues(&er);
	}while(!((er.ap_event == APPEVNT_BUTTON) &&
		 !(er.ap_value & LEFT_BUTTON)));

	if(elem) {
		Rdwrs_operation(WSTARTWRITE);

		win = findwin(win_id);

		win->tree[elem].ob_state &= ~SELECTED;

		drawwinelem(evntglbl.evid,win,&globals.screen,elem);

		Rdwrs_operation(WENDWRITE);
	};

	Appl_do_write(owner,16,&mesag);

	Evhd_release_mctrl(EVHD_APID);		
}

static void handle_mover_click(WORD win_id) {
	WORD      timeleft = 100;
  WORD      last_x = globals.mouse_x;
  WORD      last_y = globals.mouse_y;
	WINSTRUCT *win;
	EVNTREC   er;
	COMMSG    mesag;
	WORD      xyarray[10];
  
	Rdwrs_operation(WSTARTWRITE);
 
  win = findwin(win_id);
  
  win->tree[WMOVER].ob_state |= SELECTED;
  
  drawwinelem(evntglbl.evid,win,&globals.screen,WMOVER);
  
  Rdwrs_operation(WENDWRITE);
  
  while(1) {
    if(get_evntpacket(&er,timeleft) == 0) {
    	timeleft = 0;
    	break;
    };
    
    if(er.ap_event == APPEVNT_TIMER) {
    	timeleft -= (WORD)er.ap_value;
    };
    
    if(timeleft < 0) {
      break;
    };
    
    update_mousevalues(&er);
    
    if((er.ap_event == APPEVNT_BUTTON) &&
    		!(LEFT_BUTTON & er.ap_value)) {
    		break;
    };
  };
  
  if(timeleft) {
    static COMMSG	m;

		Rdwrs_operation(WSTARTREAD);
		
	  win = findwin(win_id);
  
    if(win->status & WIN_ISTOPPED) {
      m.type = WM_BOTTOM;
    }
    else {
      m.type = WM_TOPPED;
    };
    
    m.sid = 0;
    m.length = 0;
    m.msg0 = win_id;
    
    Appl_do_write(win->owner,16,&m);

		Rdwrs_operation(WENDREAD);
  }
  else {
    if(globals.realmove) {
    	Rdwrs_operation(WSTARTREAD);
    	
      win = findwin(win_id);

      mesag.type = WM_MOVED;
      mesag.sid = 0;
      mesag.length = 0;
      mesag.msg0 = win_id;
      mesag.msg1 = win->totsize.x;
      mesag.msg2 = win->totsize.y;
      mesag.msg3 = win->totsize.width;
      mesag.msg4 = win->totsize.height;

			Rdwrs_operation(WENDREAD);
      
      do {
        WORD	waittime = 200;
        
        if((last_x != globals.mouse_x) || (last_y != globals.mouse_y)) {
          mesag.msg1 += globals.mouse_x - last_x;
          mesag.msg2 += globals.mouse_y - last_y;
          
          last_x = globals.mouse_x;
          last_y = globals.mouse_y;
          
          Appl_do_write(win->owner,16,&mesag);
        };
        
        while(1) {
          if(get_evntpacket(&er,(WORD)waittime) == 0) {
            break;
          };
          
          if(er.ap_event == APPEVNT_TIMER) {
            if((waittime -= (WORD)er.ap_value) <= 0) {
              break;
            };
          }
          else {
            update_mousevalues(&er);
          };
          
          if((er.ap_event == APPEVNT_BUTTON) &&
             !(er.ap_value & LEFT_BUTTON)) {
            break;
          };
        };
      }while(!((er.ap_event == APPEVNT_BUTTON) &&
               !(er.ap_value & LEFT_BUTTON)));
    }
    else {
			Wind_beg_update(0);  /*<--???*/

			Rdwrs_operation(WSTARTREAD);

      win = findwin(win_id);

      Vdi_vsl_type(evntglbl.evid,3);
      Vdi_vswr_mode(evntglbl.evid,MD_XOR);
      
      xyarray[0] = win->totsize.x;
      xyarray[1] = win->totsize.y;
      xyarray[2] = win->totsize.x
        + win->totsize.width - 1;
      xyarray[3] = xyarray[1];
      xyarray[4] = xyarray[2];
      xyarray[5] = win->totsize.y
        + win->totsize.height - 1;
      xyarray[6] = xyarray[0];
      xyarray[7] = xyarray[5];
      xyarray[8] = xyarray[0];
      xyarray[9] = xyarray[1];

      do {
        xyarray[0] += globals.mouse_x - last_x;
        xyarray[1] += globals.mouse_y - last_y;
        xyarray[2] += globals.mouse_x - last_x;
        xyarray[3] = xyarray[1];
        xyarray[4] = xyarray[2];
        xyarray[5] += globals.mouse_y - last_y;
        xyarray[6] = xyarray[0];
        xyarray[7] = xyarray[5];
        xyarray[8] = xyarray[0];
        xyarray[9] = xyarray[1];
        
        last_x = globals.mouse_x;
        last_y = globals.mouse_y;
        
        Vdi_v_hide_c(evntglbl.evid);
        Vdi_v_pline(evntglbl.evid,5,xyarray);
        Vdi_v_show_c(evntglbl.evid,1);
        
        get_evntpacket(&er,0);
        update_mousevalues(&er);
        
        Vdi_v_hide_c(evntglbl.evid);
        Vdi_v_pline(evntglbl.evid,5,xyarray);
        Vdi_v_show_c(evntglbl.evid,1);
      }while(!((er.ap_event == APPEVNT_BUTTON) &&
               !(er.ap_value & LEFT_BUTTON)));
      
      
      mesag.type = WM_MOVED;
      mesag.sid = 0;
      mesag.length = 0;
      mesag.msg0 = win_id;
      mesag.msg1 = xyarray[0];
      mesag.msg2 = xyarray[1];
      mesag.msg3 = win->totsize.width;
      mesag.msg4 = win->totsize.height;
      
      Appl_do_write(win->owner,16,&mesag);

			Rdwrs_operation(WENDREAD);
			
      Wind_end_update(0);  /*<--???*/
    };
  };
	
	Rdwrs_operation(WSTARTREAD);
  
  win = findwin(win_id);
  
  win->tree[WMOVER].ob_state &= ~SELECTED;
  drawwinelem(evntglbl.evid,win,&globals.screen,WMOVER);
  
	Rdwrs_operation(WENDREAD);
}

static void handle_slider_click(WORD win_id,WORD elem) {
  WORD      last_x = globals.mouse_x;
  WORD      last_y = globals.mouse_y;
  WORD      dx,dy;
	WINSTRUCT *win;
	EVNTREC   er;
	COMMSG    mesag;
	WORD      xyarray[10];
	RECT      elemrect,bgrect;
	WORD      bg;
  
  if(elem == WVSLIDER) {
  	bg = WVSB;
  }
  else {
  	bg = WHSB;
  };
  
	Rdwrs_operation(WSTARTWRITE);
 
  win = findwin(win_id);
  
  win->tree[elem].ob_state |= SELECTED;
  
  drawwinelem(evntglbl.evid,win,&globals.screen,elem);
 
 	Objc_do_offset(win->tree,elem,(WORD *)&elemrect);
 	elemrect.width = win->tree[elem].ob_width;
 	elemrect.height = win->tree[elem].ob_height;
  
  dx = last_x - elemrect.x;
  dy = last_y - elemrect.y;
  
 	Objc_do_offset(win->tree,bg,(WORD *)&bgrect);
 	bgrect.width = win->tree[bg].ob_width;
 	bgrect.height = win->tree[bg].ob_height;
  
  Rdwrs_operation(WENDWRITE);
  
	Wind_beg_update(0);  /*<--???*/

	Rdwrs_operation(WSTARTREAD);

  win = findwin(win_id);

  Vdi_vsl_type(evntglbl.evid,3);
  Vdi_vswr_mode(evntglbl.evid,MD_XOR);
  
  xyarray[0] = elemrect.x;
  xyarray[1] = elemrect.y;
  xyarray[2] = elemrect.x
				       + elemrect.width - 1;
  xyarray[3] = xyarray[1];
  xyarray[4] = xyarray[2];
  xyarray[5] = elemrect.y
				       + elemrect.height - 1;
  xyarray[6] = xyarray[0];
  xyarray[7] = xyarray[5];
  xyarray[8] = xyarray[0];
  xyarray[9] = xyarray[1];

  Vdi_v_hide_c(evntglbl.evid);
  Vdi_v_pline(evntglbl.evid,5,xyarray);
  Vdi_v_show_c(evntglbl.evid,1);
      
  while(1) {
  	WORD new_x,new_y;
  	
    get_evntpacket(&er,0);
    update_mousevalues(&er);
    
    if((er.ap_event == APPEVNT_BUTTON) &&
           !(er.ap_value & LEFT_BUTTON)) {
    	break;
    };  
  	
  	if(elem == WHSLIDER) {
  		new_y = last_y;
  		
  		if((globals.mouse_x - dx + elemrect.width) > (bgrect.x + bgrect.width)) {
  			new_x = bgrect.x + bgrect.width - elemrect.width + dx;
  		}
  		else if((globals.mouse_x - dx) < bgrect.x) {
  			new_x = bgrect.x + dx;
  		}
  		else {
  			new_x = globals.mouse_x;
  		};

			if(new_x != last_x) {
	      Vdi_v_hide_c(evntglbl.evid);
	      Vdi_v_pline(evntglbl.evid,5,xyarray);
	      Vdi_v_show_c(evntglbl.evid,1);
      
        xyarray[0] += new_x - last_x;
        xyarray[2] += new_x - last_x;
	      xyarray[4] = xyarray[2];
  	    xyarray[6] = xyarray[0];
    	  xyarray[8] = xyarray[0];

	      Vdi_v_hide_c(evntglbl.evid);
	      Vdi_v_pline(evntglbl.evid,5,xyarray);
	      Vdi_v_show_c(evntglbl.evid,1);
    	};
  	}
  	else {
  		new_x = last_x;
  		
  		if((globals.mouse_y - dy + elemrect.height) > (bgrect.y + bgrect.height)) {
  			new_y = bgrect.y + bgrect.height - elemrect.height + dy;
  		}
  		else if((globals.mouse_y - dy) < bgrect.y) {
  			new_y = bgrect.y + dy;
  		}
  		else {
  			new_y = globals.mouse_y;
  		};

  		if(new_y != last_y) {
        Vdi_v_hide_c(evntglbl.evid);
		    Vdi_v_pline(evntglbl.evid,5,xyarray);
        Vdi_v_show_c(evntglbl.evid,1);
      
      	xyarray[1] += new_y - last_y;
        xyarray[3] = xyarray[1];
	      xyarray[5] += new_y - last_y;
  	    xyarray[7] = xyarray[5];
    	  xyarray[9] = xyarray[1];

        Vdi_v_hide_c(evntglbl.evid);
		    Vdi_v_pline(evntglbl.evid,5,xyarray);
    		Vdi_v_show_c(evntglbl.evid,1);
      };
  	};
    
    last_x = new_x;
    last_y = new_y;
  };
  
  Vdi_v_hide_c(evntglbl.evid);
  Vdi_v_pline(evntglbl.evid,5,xyarray);
  Vdi_v_show_c(evntglbl.evid,1);
      
	if(elem == WHSLIDER) {
    mesag.type = WM_HSLID;
    if(bgrect.width == elemrect.width) {
    	mesag.msg1 = 0;
    }
    else {
			mesag.msg1 = (WORD)((((LONG)last_x - (LONG)dx - (LONG)bgrect.x) * (LONG)1000L) / 
										((LONG)bgrect.width - (LONG)elemrect.width));
		};
	}
	else {
    mesag.type = WM_VSLID;
    if(bgrect.height == elemrect.height) {
    	mesag.msg1 = 0;
    }
    else {
			mesag.msg1 = (WORD)((((LONG)last_y - (LONG)dy - (LONG)bgrect.y) * (LONG)1000L) / 
										((LONG)bgrect.height - (LONG)elemrect.height));
		};
	};
	      
  mesag.sid = 0;
  mesag.length = 0;
  mesag.msg0 = win_id;
  
  Appl_do_write(win->owner,16,&mesag);

	Rdwrs_operation(WENDREAD);
	
  Wind_end_update(0);

	Rdwrs_operation(WSTARTREAD);
  
  win = findwin(win_id);
  
  win->tree[elem].ob_state &= ~SELECTED;
  drawwinelem(evntglbl.evid,win,&globals.screen,elem);
  
	Rdwrs_operation(WENDREAD);
}

static void	handle_button(WORD newbutton) {
  EVNTREC	er;
  WORD	changed = newbutton ^ globals.mouse_button;
  
  globals.mouse_button = newbutton;
  
  if(mctrl.apid > -1) {
    EVNTREC	e;
	      
    e.ap_event = APPEVNT_BUTTON;
    e.ap_value = globals.mouse_button | (1L << 16);
    Appl_put_event((WORD)mctrl.apid,&e);
  }
  else {
	  if((globals.mouse_button & LEFT_BUTTON & changed)) {
	    WORD	win_id;
    
	    WINSTRUCT	*win;
    
			Rdwrs_operation(WSTARTREAD);
	    
	    win_id = winoncoord(globals.mouse_x,globals.mouse_y);
    
			win = findwin(win_id);
    
    	if(win->status & WIN_DESKTOP) {
      	EVNTREC	e;

      	e.ap_event = APPEVNT_BUTTON;
      	e.ap_value = globals.mouse_button | (1L << 16);
      	Appl_put_event(DESK_OWNER,&e);

				Rdwrs_operation(WENDREAD);
    	}
    	else if((win->status & WIN_DIAL)
      	|| (win->status & WIN_MENU) ||
       	((win->status & WIN_UNTOPPABLE)
				&& (globals.mouse_x >= win->worksize.x)
				&& (globals.mouse_x < win->worksize.x + win->worksize.width)
				&& (globals.mouse_y >= win->worksize.y)
				&& (globals.mouse_y < win->worksize.y + win->worksize.height))) {
      	EVNTREC	e;
      
      	e.ap_event = APPEVNT_BUTTON;
      	e.ap_value = globals.mouse_button | (1L << 16);
      	Appl_put_event(win->owner,&e);

				Rdwrs_operation(WENDREAD);
    	}
    	else {	
      	WORD	obj = Objc_do_find(win->tree,0,9,globals.mouse_x,globals.mouse_y,0);
      	WORD	owner = win->owner;
				COMMSG	mesag;				
				WORD	xyarray[10];
      
				Rdwrs_operation(WENDREAD);
      	
      	switch(obj) {		
	      case	WCLOSER	:
					Rdwrs_operation(WSTARTREAD);
		
					win = findwin(win_id);
					win->tree[WCLOSER].ob_state |= SELECTED;
					drawwinelem(evntglbl.evid,win,&globals.screen,WCLOSER);

					Rdwrs_operation(WENDREAD);
		
					do {
					  get_evntpacket(&er,0);
					  update_mousevalues(&er);
					}while(!((er.ap_event == APPEVNT_BUTTON) &&
						 !(er.ap_value & LEFT_BUTTON)));
		
					Rdwrs_operation(WSTARTREAD);
		
					win = findwin(win_id);
		
					win->tree[WCLOSER].ob_state &= ~SELECTED;
		
					drawwinelem(evntglbl.evid,win,&globals.screen,WCLOSER);
		
					Rdwrs_operation(WENDREAD);
		
					mesag.type = WM_CLOSED;
					mesag.sid = 0;
					mesag.length = 0;
					mesag.msg0 = win_id;
					Appl_do_write(owner,16,&mesag);
		
					break;
		
	      case WFULLER:
					Rdwrs_operation(WSTARTREAD);
		
					win = findwin(win_id);
		
					win->tree[WFULLER].ob_state |= SELECTED;
		
					drawwinelem(evntglbl.evid,win,&globals.screen,WFULLER);
		
					Rdwrs_operation(WENDREAD);
		
					do {
					  get_evntpacket(&er,0);
					  update_mousevalues(&er);
					}while(!((er.ap_event == APPEVNT_BUTTON) &&
						 !(er.ap_value & LEFT_BUTTON)));
		
					Rdwrs_operation(WSTARTREAD);
		
					win = findwin(win_id);
		
					win->tree[WFULLER].ob_state &= ~SELECTED;
		
					drawwinelem(evntglbl.evid,win,&globals.screen,WFULLER);

					Rdwrs_operation(WENDREAD);

					mesag.type = WM_FULLED;
					mesag.sid = 0;
					mesag.length = 0;
					mesag.msg0 = win_id;
					Appl_do_write(owner,16,&mesag);
		
					break;
		
	      case	WSIZER	:
					Rdwrs_operation(WSTARTWRITE);

					win = findwin(win_id);
		
					win->tree[WSIZER].ob_state |= SELECTED;
		
					drawwinelem(evntglbl.evid,win,&globals.screen,WSIZER);
		
					if(globals.realmove) {
						WORD	last_x;
						WORD	last_y;
					  WORD	offsx = globals.mouse_x - win->totsize.x - win->totsize.width;
					  WORD	offsy = globals.mouse_y - win->totsize.y - win->totsize.height;
		  
					  mesag.type = WM_SIZED;
					  mesag.sid = 0;
					  mesag.length = 0;
					  mesag.msg0 = win_id;
					  mesag.msg1 = win->totsize.x;
					  mesag.msg2 = win->totsize.y;

						Rdwrs_operation(WENDWRITE);

						last_x = globals.mouse_x;
						last_y = globals.mouse_y;
		  
				    do {
	    			  LONG	waittime = 200;
	      
				      if((last_x != globals.mouse_x) || (last_y != globals.mouse_y)) {
				        mesag.msg3 = globals.mouse_x - win->totsize.x - offsx;
								mesag.msg4 = globals.mouse_y - win->totsize.y - offsy;
	        
				        last_x = globals.mouse_x;
				        last_y = globals.mouse_y;
	        
				        Appl_do_write(win->owner,16,&mesag);
				      };
	      
				      while(1) {
				        if(get_evntpacket(&er,(WORD)waittime) == 0) {
				          break;
				        };
	        
				        if(er.ap_event == APPEVNT_TIMER) {
				          if((waittime -= er.ap_value) <= 0) {
				            break;
				          };
				        }
				        else {
				          update_mousevalues(&er);
				        };
	        
				        if((er.ap_event == APPEVNT_BUTTON) &&
				           !(er.ap_value & LEFT_BUTTON)) {
				          break;
				        };
				      };
				    }while(!((er.ap_event == APPEVNT_BUTTON) &&
				             !(er.ap_value & LEFT_BUTTON)));
				 	}
					else {
					  Vdi_vsl_type(evntglbl.evid,3);
					  Vdi_vswr_mode(evntglbl.evid,MD_XOR);
					  
					  xyarray[0] = win->totsize.x;
					  xyarray[1] = win->totsize.y;
					  xyarray[3] = xyarray[1];
					  xyarray[6] = xyarray[0];
					  xyarray[8] = xyarray[0];
					  xyarray[9] = xyarray[1];
				
						Wind_beg_update(0);
					  
					  do {
					    xyarray[2] = globals.mouse_x;
					    xyarray[4] = xyarray[2];
					    xyarray[5] = globals.mouse_y;
					    xyarray[7] = xyarray[5];
					    
					    Vdi_v_hide_c(evntglbl.evid);
					    Vdi_v_pline(evntglbl.evid,5,xyarray);
	            Vdi_v_show_c(evntglbl.evid,1);
	            
	            get_evntpacket(&er,0);
	            update_mousevalues(&er);
	            
	            Vdi_v_hide_c(evntglbl.evid);
	            Vdi_v_pline(evntglbl.evid,5,xyarray);
	            Vdi_v_show_c(evntglbl.evid,1);
	          }while(!((er.ap_event == APPEVNT_BUTTON) &&
	                   !(er.ap_value & LEFT_BUTTON)));
	          
	          Wind_end_update(0);
	          
	          mesag.type = WM_SIZED;
	          mesag.sid = 0;
	          mesag.length = 0;
	          mesag.msg0 = win_id;
	          mesag.msg1 = win->totsize.x;
	          mesag.msg2 = win->totsize.y;
	          mesag.msg3 = globals.mouse_x - win->totsize.x - 1;
	          mesag.msg4 = globals.mouse_y - win->totsize.y - 1;

						Rdwrs_operation(WENDWRITE);

	          Appl_do_write(owner,16,&mesag);
	        };

					Rdwrs_operation(WSTARTREAD);
	        
	        win = findwin(win_id);
	        
	        win->tree[WSIZER].ob_state &= ~SELECTED;
	        drawwinelem(evntglbl.evid,win,&globals.screen,WSIZER);

					Rdwrs_operation(WENDREAD);
	        break;
	        
	      case	WMOVER	:
	      	handle_mover_click(win_id);
	        break;
	        
	      case WLEFT:
					handle_arrow_click(win_id,WLEFT,WA_LFLINE);
					break;
		
	      case WRIGHT:
					handle_arrow_click(win_id,WRIGHT,WA_RTLINE);
					break;
		
	      case WUP:
					handle_arrow_click(win_id,WUP,WA_UPLINE);
					break;
		
	      case WDOWN:
					handle_arrow_click(win_id,WDOWN,WA_DNLINE);
					break;
		
	      case WHSB:
	      	{
	      		WORD      xy[2];
	      		WINSTRUCT *win;
	      		
						Rdwrs_operation(WSTARTREAD);
						
						if((win = findwin(win_id))) {
			      	Objc_do_offset(win->tree,WHSLIDER,xy);
			      };
		      	
						Rdwrs_operation(WENDREAD);
	
						if(globals.mouse_x > xy[0]) {
							handle_arrow_click(win_id,0,WA_RTPAGE);
						}
						else {
							handle_arrow_click(win_id,0,WA_LFPAGE);
						};
					};
					break;
		
	      case WVSB:
	      	{
	      		WORD      xy[2];
	      		WINSTRUCT *win;
	      		
						Rdwrs_operation(WSTARTREAD);
						
						if((win = findwin(win_id))) {
			      	Objc_do_offset(win->tree,WVSLIDER,xy);
			      };
		      	
						Rdwrs_operation(WENDREAD);
	
						if(globals.mouse_y > xy[1]) {
							handle_arrow_click(win_id,0,WA_DNPAGE);
						}
						else {
							handle_arrow_click(win_id,0,WA_UPPAGE);
						};
					};
					break;
		
	      case WHSLIDER:
	      	handle_slider_click(win_id,WHSLIDER);
	        break;
	        
	      case WVSLIDER:
	      	handle_slider_click(win_id,WVSLIDER);
	        break;
	        
	      default:
	        if(win_id > 0) {
	          COMMSG	m;
	          
	          WINSTRUCT	*win;
	          
						Rdwrs_operation(WSTARTREAD);
	          
	          win = findwin(win_id);
	          	          
	          if(win && (win != globals.win_vis->win)) {
			        do {
	  		        get_evntpacket(&er,0);
	      		    update_mousevalues(&er);
			        }while(!((er.ap_event == APPEVNT_BUTTON) &&
	                 !(er.ap_value & LEFT_BUTTON)));
	        
	            m.type = WM_TOPPED;
	            m.sid = 0;
	            m.length = 0;
	            m.msg0 = win_id;
	            
	            Appl_do_write(win->owner,16,&m);
	          }
	          else {
			      	er.ap_event = APPEVNT_BUTTON;
			      	er.ap_value = (LONG)globals.mouse_button | (1L << 16);
			      	Appl_put_event(win->owner,&er);
						};

						Rdwrs_operation(WENDREAD);
	        };
	      };
	    };
	  }
	  else {
	    WORD	win_id;
	    
	    WINSTRUCT	*win;
	    
			Rdwrs_operation(WSTARTREAD);
	    
	    win_id = winoncoord(globals.mouse_x,globals.mouse_y);
	    win = findwin(win_id);
 
	    if(win->status & WIN_DESKTOP) {
	      EVNTREC	e;
	      
	      e.ap_event = APPEVNT_BUTTON;
	      e.ap_value = globals.mouse_button | (1L << 16);
	      Appl_put_event(DESK_OWNER,&e);
	    }
	    else if(Misc_inside(&win->worksize,globals.mouse_x,
				globals.mouse_y)) {
	      EVNTREC	e;
	      
	      e.ap_event = APPEVNT_BUTTON;
	      e.ap_value = globals.mouse_button | (1L << 16);
	      Appl_put_event(win->owner,&e);
	    };

	    Rdwrs_operation(WENDREAD);
	  };
	};
}

static void check_rectevents(WORD x,WORD y) {
  REVENTLIST	**rl;
  
  struct {
    EVNTREC       head;
    EVNTREC_MOUSE tail;
  }m;
  
  if(!evntglbl.eventlist) { /* This speeds things up when there is nothing in */
    return;                 /* the list */
  };
    
  Psemaphore(2,MOUSESEM,-1);
  
  rl = &evntglbl.eventlist;
  
  while(*rl) {
  	if((mctrl.apid == -1) || (mctrl.apid == (*rl)->event.apid)) {
			WORD    sendflag = 0;
			AP_INFO *ai;

	    if((*rl)->event.flag1 != -1) {
	    	WORD insideflag = Misc_inside(&(*rl)->event.r1,x,y);
	    	
	    	if((insideflag && ((*rl)->event.flag1 == MO_ENTER)) ||
	    		((!insideflag) && ((*rl)->event.flag1 == MO_LEAVE))) {	
					m.head.ap_event = MO_RECT1;
					m.head.ap_value = (*rl)->event.apid;
					m.tail.mx = x;
					m.tail.my = y;
					m.tail.buttons = globals.mouse_button;
					/*m.tail.kstate = TODO <--*/
	
					sendflag = 1;	
		    };
		  }
	    else if((*rl)->event.flag2 != -1) {
	    	WORD insideflag = Misc_inside(&(*rl)->event.r2,x,y);
	    	
	    	if((insideflag && ((*rl)->event.flag2 == MO_ENTER)) ||
	    		((!insideflag) && ((*rl)->event.flag2 == MO_LEAVE))) {
					m.head.ap_event = MO_RECT2;
					m.head.ap_value = (*rl)->event.apid;
					m.tail.mx = x;
					m.tail.my = y;
					m.tail.buttons = globals.mouse_button;
					/*m.tail.kstate = TODO <--*/
	
					sendflag = 1;
		    };
		  };

			if(sendflag) {
				REVENTLIST *tmp = *rl;

				Rdwrs_operation(ASTARTREAD);
				
				ai = Appl_internal_info(tmp->event.apid);
	
				if(ai) {
	  			LONG pnr = Fopen(ai->eventname,O_WRONLY);
	
	  			Fwrite((WORD)pnr,sizeof(EVNTREC) + sizeof(EVNTREC_MOUSE),&m);
	  			Fclose((WORD)pnr);
				}
				else {
	  			fprintf(stderr,"%s: Line %d: "
		  			"Couldn't find application description!\r\n",
		  			__FILE__,__LINE__);
				};
				
				Rdwrs_operation(AENDREAD);
	
				/* Make sure that we only send one message =>
	   			remove the entry*/
	
				*rl = (*rl)->next;
				Mfree(tmp);
			};
				    
	    if(!(*rl))
	      break;
	  };	    
    rl = &(*rl)->next;
  };
  
  Psemaphore(3,MOUSESEM,-1);
}

static mouseproc(LONG arg) {
  WORD	work_in[] = {1,7,1,1,1,1,1,1,1,1,2};
  WORD	work_out[57];
  EVNTREC	er;

	NOT_USED(arg);

  evntglbl.mousefd = (WORD)Fopen("u:\\dev\\aesmouse",0);
  evntglbl.lasttime = globals.time;

  
  /* open up a vdi workstation to use in the process */
  
  evntglbl.evid = globals.vid;
  Vdi_v_opnvwk(work_in,&evntglbl.evid,work_out);
  
  while(1) {
    get_evntpacket(&er,0);

    switch((WORD)er.ap_event) {		
    case	APPEVNT_BUTTON	:
      handle_button((WORD)er.ap_value);
      break;	
      
    case	APPEVNT_MOUSE	:
      globals.mouse_x = ((WORD *)&er.ap_value)[1];
      globals.mouse_y = ((WORD *)&er.ap_value)[0];

		if((mctrl.apid > -1) && (mctrl.mode == MCTRL_COMPLETE)) {
		    Appl_put_event((WORD)mctrl.apid,&er);
  		};

      check_rectevents(globals.mouse_x,globals.mouse_y);
      break;
      
    case	APPEVNT_KEYBOARD :
      Appl_put_event(TOP_APPL,&er);
      break;
    };
  };
}

/****************************************************************************
 * Public functions                                                         *
 ****************************************************************************/

/****************************************************************************
 * Evhd_init_module                                                         *
 *  Initiate event processing module.                                       *
 ****************************************************************************/
void                   /*                                                   */
Evhd_init_module(void) /*                                                   */
/****************************************************************************/
{
  evntglbl.mouseid = (WORD)newfork(mouseproc,0,"MouseLoop");
  
  /* Initialize internal mouse variables */
  
  Vdi_vq_mouse(globals.vid,&globals.mouse_button,&globals.mouse_x,&globals.mouse_y);
  
  /* Create and get the mouse event semaphore */
  
  Psemaphore(0,MOUSESEM,-1);
  
  /* Release it */
  
  Psemaphore(3,MOUSESEM,-1);

  Psemaphore(0,MCTRL_SEM,-1);
  Psemaphore(3,MCTRL_SEM,-1);
}

/****************************************************************************
 * Evhd_exit_module                                                         *
 *  Shutdown event processing module.                                       *
 ****************************************************************************/
void                   /*                                                   */
Evhd_exit_module(void) /*                                                   */
/****************************************************************************/
{
  /* Get the mouse event semaphore */
  
  Psemaphore(2,MOUSESEM,-1);
  
  /* Destroy it */
  
  Psemaphore(1,MOUSESEM,-1);
  
  Psemaphore(2,MCTRL_SEM,-1);
  Psemaphore(1,MCTRL_SEM,-1);

  Pkill(evntglbl.mouseid,SIGKILL);
}

/****************************************************************************
 * Evhd_make_rectevent                                                      *
 *  Start reporting of mouse events.                                        *
 ****************************************************************************/
void                   /*                                                   */
Evhd_make_rectevent(   /*                                                   */
RECTEVENT *re)         /* Description of events that should be reported.    */
/****************************************************************************/
{
  REVENTLIST *rl = (REVENTLIST *)Mxalloc(sizeof(REVENTLIST),GLOBALMEM);
  
  Psemaphore(2,MOUSESEM,-1);
  
  rl->event = *re;
  rl->next = evntglbl.eventlist;
  evntglbl.eventlist = rl;
  
  Psemaphore(3,MOUSESEM,-1);
  
  check_rectevents(globals.mouse_x,globals.mouse_y);
}

/****************************************************************************
 * Evhd_kill_rectevent                                                      *
 *  End reporting of mouse events.                                          *
 ****************************************************************************/
void                   /*                                                   */
Evhd_kill_rectevent(   /*                                                   */
WORD apid)             /* Application id to end reporting to.               */
/****************************************************************************/
{
  REVENTLIST	**rl;
  
  Psemaphore(2,MOUSESEM,-1);
  
  rl = &evntglbl.eventlist;
  
  while(*rl) {
    if((*rl)->event.apid == apid) {
      REVENTLIST	*tmp = *rl;
      
      *rl = tmp->next;
      
      Mfree(tmp);
      break;
    };
    
    rl = &(*rl)->next;
  };
  
  Psemaphore(3,MOUSESEM,-1);
}

/****************************************************************************
 * Evhd_gain_mctrl                                                          *
 *  Gain exclusive right to mouse events.                                   *
 ****************************************************************************/
WORD              /*                                                        */
Evhd_gain_mctrl(  /*                                                        */
WORD apid,        /* Application to gain exclusive right.                   */
WORD mode)        /* MCTRL_COMPLETE or MCTRL_BUTTON.                        */
/****************************************************************************/
{
	WORD r;

	if((mctrl.count > 0) && (mctrl.apid != apid))
		return 0;
	
	r = Rdwrs_get_sem(apid, &mctrl);
	if(r)
		mctrl.mode = mode;

	return r;
}

/****************************************************************************
 * Evhd_release_mctrl                                                       *
 *  Release right to mouse events.                                          *
 ****************************************************************************/
WORD                 /*                                                     */
Evhd_release_mctrl(  /*                                                     */
WORD apid)           /* Application to release exclusive right.             */
/****************************************************************************/
{
	return Rdwrs_rel_sem( apid, &mctrl);
}
