/****************************************************************************

 Module
  objc.c
  
 Description
  Resource object routines in oAESis.
  
 Author(s)
 	cg (Christer Gustavsson <d2cg@dtek.chalmers.se>)
 	
 Revision history
 
  951229 cg
   Added standard header.
   get_cliprect() renamed to Objc_area_needed().
 
 Copyright notice
  The copyright to the program code herein belongs to the authors. It may
  be freely duplicated and distributed without fee, but not charged for.
 
 ****************************************************************************/

/****************************************************************************
 * Used interfaces                                                          *
 ****************************************************************************/

#define	__TCC_COMPAT__


#include <gemfast.h>
#include <mintbind.h>
#include <stdio.h>

#include "appl.h"
#include "global.h"
#include "misc.h"
#include "objc.h"
#include "types.h"
#include "vdi.h"

/****************************************************************************
 * Macros                                                                   *
 ****************************************************************************/

#define	SWRM_TRANS		0
#define	SWRM_REPLACE	1
#define	SWRM_INVTRANS	4
#define BUTTONFRAME   2
#define DEFBUTFRAME   3
#define OUTLINESIZE   3

/****************************************************************************
 * Typedefs of module global interest                                       *
 ****************************************************************************/

typedef struct tiext {
	unsigned	framecol    : 4;
	unsigned	textcol     : 4;
	unsigned	textmode    : 1;
	unsigned	fillpattern : 3;
	unsigned	interiorcol : 4;
}TIEXT;

/****************************************************************************
 * Module global variables                                                  *
 ****************************************************************************/

static WORD	maxnumcols;
static WORD num_planes;

static WORD	largeth = 16;
static WORD largetw = 8;
static WORD	smallth = 8;
static WORD	textyoffs = 8;

/****************************************************************************
 * Local functions (use static!)                                            *
 ****************************************************************************/

static WORD	invertcolor(WORD color) {
	switch(color) {
		case	BLACK:
			return	WHITE;
		case	WHITE:
			return	BLACK;
		case	BLUE:
			return	LYELLOW;
		case	LYELLOW:
			return	BLUE;
		case	RED:
			return	LCYAN;
		case	LCYAN:
			return	RED;
		case	GREEN:
			return	LMAGENTA;
		case	LMAGENTA:
			return	GREEN;
		case	CYAN:
			return	LRED;
		case	LRED:
			return	CYAN;
		case	YELLOW:
			return	LBLUE;
		case	LBLUE:
			return	YELLOW;
		case	MAGENTA:
			return	LGREEN;
		case	LGREEN:
			return	MAGENTA;
		case	LWHITE:
			return	LBLACK;
		case	LBLACK:
			return	LWHITE;
		default:
			return	BLACK;
	};
}

static void	setfilltype(WORD vid,WORD fill) {
	switch(fill) {
		case 0	:
			Vdi_vsf_interior(vid,0);
			break;
		case 7	:
			Vdi_vsf_interior(vid,1);
			break;
		default	:
			Vdi_vsf_interior(vid,2);
			Vdi_vsf_style(vid,fill);
			break;
	}
}

static void	settxtalign(WORD vid,WORD alignment) {
	Vdi_vst_alignment(vid,alignment,3,&alignment,&alignment);
}

static void	setwritemode(WORD vid,WORD mode) {
	switch(mode) {
		case	SWRM_TRANS		:	/*transparent*/
			Vdi_vswr_mode(vid,MD_TRANS);
			break;
		case	SWRM_REPLACE	:	/*replace*/
			Vdi_vswr_mode(vid,MD_REPLACE);
			break;
			
		case	SWRM_INVTRANS	:	/*0x0004*/
			Vdi_vswr_mode(vid,MD_ERASE);
			break;
	};
}

static void	settextsize(WORD vid,WORD size) {
	switch(size) {
	case	3:	/*large*/
		Vdi_vst_font( vid, globals.fnt_regul_id);
		Vdi_vst_height(vid,globals.fnt_regul_sz,&size,&size,&size,&size);
					
		textyoffs = largeth >> 1;
		break;
	
	case	5:	/*small*/
		Vdi_vst_font( vid, globals.fnt_small_id);
		Vdi_vst_height(vid,globals.fnt_small_sz,&size,&size,&size,&size);
		
		textyoffs = smallth >> 1;	
	};
}

static void	drawtext(WORD vid,char *text,RECT *size,WORD alignment,WORD color
	,WORD textsize,WORD writemode,WORD state) {
	WORD x;
	WORD y;
	WORD txteff = 0;
	
	settextsize(vid,textsize);

	y = size->y + (size->height >> 1) + textyoffs - 1;

	switch(alignment) {
		case	0	:	/*left*/
			x = size->x;
			settxtalign(vid,0);
			break;
		case	1	:	/*right*/
			x = size->x + size->width - 1;
			settxtalign(vid,2);
			break;
		case	2	:	/*center*/
			x = size->x + (size->width >> 1);
			settxtalign(vid,1);
			break;
	};
	
	if(state & DISABLED) {
		txteff |= LIGHT;
	};
	
	Vdi_vst_effects(vid,txteff);
	
	if(state & SELECTED) {
		setwritemode(vid,writemode);
		
		Vdi_vst_color(vid,invertcolor(color));
		Vdi_v_gtext(vid,x,y,text);
		
		if(writemode == SWRM_REPLACE)
		{
			setwritemode(vid,SWRM_INVTRANS);
			Vdi_vst_color(vid,BLACK);
			Vdi_v_gtext(vid,x,y,text);
		};
	}
	else {
		setwritemode(vid,writemode);
		Vdi_vst_color(vid,color);
		Vdi_v_gtext(vid,x,y,text);
	};
}

static void	drawframe(WORD vid,RECT *r,WORD framesize) {
	register WORD	incr;
	WORD	size[10];
	register WORD	thick;
	
	if(framesize) {
		if(framesize > 0) {
			thick = framesize;
			incr = 1;
			size[0] = r->x;
			size[1] = r->y;
			size[2] = size[0] + r->width - 1;
			size[3] = size[1];
			size[4] = size[2];
			size[5] = size[1] + r->height - 1;
			size[6] = size[0];
			size[7] = size[5];
			size[8] = size[0];
			size[9] = size[1];			
		}
		else {
			thick =  -framesize;
			incr = -1;
			size[0] = r->x - 1;
			size[1] = r->y - 1;
			size[2] = r->x + r->width;
			size[3] = size[1];
			size[4] = size[2];
			size[5] = r->y + r->height;
			size[6] = size[0];
			size[7] = size[5];
			size[8] = size[0];
			size[9] = size[1];			
		};
			
		while(thick > 0) {
			Vdi_v_pline(vid,5,size);
			thick--;
			if(!thick)
				break;
				
				
			size[0] += incr;
			size[1] += incr;
			size[2] -= incr;
			size[3] += incr;
			size[4] -= incr;
			size[5] -= incr;
			size[6] += incr;
			size[7] -= incr;
			size[8] += incr;
			size[9] += incr;		
		};
	};
}

static void	drawrectfi(WORD vid,RECT *r,WORD framesize) {
	register WORD	incr;
	WORD	size[10];
	register WORD	thick;
	
	size[0] = r->x;
	size[1] = r->y;
	size[2] = r->width + size[0] - 1;
	size[3] = r->height + size[1] - 1;

	Vdi_vr_recfl(vid,size);

	Vdi_vsl_type(vid,SOLID);

	if(framesize) {
		if(framesize > 0) {
			thick = framesize;
			incr = 1;
		}
		else {
			thick = -framesize;
			incr = -1;
			size[0]--;
			size[1]--;
			size[2]++;
			size[3]++;
		};
		
		size[5] = size[3];
		size[3] = size[1];
		size[4] = size[2];
		size[6] = size[0];
		size[7] = size[5];
		size[8] = size[0];
		size[9] = size[1];
	
		while(thick > 0) {
			Vdi_v_pline(vid,5,size);
			thick--;
			if(!thick)
				break;
				
				
			size[0] += incr;
			size[1] += incr;
			size[2] -= incr;
			size[3] += incr;
			size[4] -= incr;
			size[5] -= incr;
			size[6] += incr;
			size[7] -= incr;
			size[8] += incr;
			size[9] += incr;		
		};
	};
}

static void drawimage(WORD vid,WORD x,WORD y,BITBLK *bb) {
	MFDB	d,s;

	WORD	xyarray[8];
	WORD	color[2];
	
	xyarray[0] = 0;
	xyarray[1] = 0;
	xyarray[2] = (bb->bi_wb << 3) - 1;
	xyarray[3] = bb->bi_hl - 1;
	xyarray[4] = x + bb->bi_x;
	xyarray[5] = y + bb->bi_y;
	xyarray[6] = x + bb->bi_x + (bb->bi_wb << 3) - 1;
	xyarray[7] = y + bb->bi_y + bb->bi_hl - 1;

	(LONG)s.fd_addr = (LONG)bb->bi_pdata;
	s.fd_w = (bb->bi_wb << 3);
	s.fd_h = bb->bi_hl;
	s.fd_wdwidth = (bb->bi_wb >> 1);
	s.fd_stand = 0;
	s.fd_nplanes = 1;

	(LONG)d.fd_addr = 0L;

	color[0] = bb->bi_color;
	
	Vdi_vrt_cpyfm(vid,MD_TRANS,xyarray,&s,&d,color);
}

static void drawicon(WORD vid,WORD x,WORD y,ICONBLK *ib) {
	char	ch[2];
	
	MFDB	d,s;

	WORD	xyarray[8];
	WORD	color[2];
	
	RECT	r;
	
	xyarray[0] = 0;
	xyarray[1] = 0;
	xyarray[2] = ib->ib_wicon - 1;
	xyarray[3] = ib->ib_hicon - 1;
	xyarray[4] = x + ib->ib_xicon;
	xyarray[5] = y + ib->ib_yicon;
	xyarray[6] = x + ib->ib_xicon + ib->ib_wicon - 1;
	xyarray[7] = y + ib->ib_yicon + ib->ib_hicon - 1;

	(LONG)s.fd_addr = (LONG)ib->ib_pmask;
	s.fd_w = ib->ib_wicon;
	s.fd_h = ib->ib_hicon;
	s.fd_wdwidth = ((ib->ib_wicon +15) >> 4);
	s.fd_stand = 0;
	s.fd_nplanes = 1;

	(LONG)d.fd_addr = 0L;

	color[0] = (ib->ib_char & 0x0f00) >> 8;
	
	Vdi_vrt_cpyfm(vid,MD_TRANS,xyarray,&s,&d,color);

	(LONG)s.fd_addr = (LONG)ib->ib_pdata;
	color[0] = ib->ib_char >> 12;
	Vdi_vrt_cpyfm(vid,MD_TRANS,xyarray,&s,&d,color);

	r.x = x + ib->ib_xtext;
	r.y = y + ib->ib_ytext;
	r.width = ib->ib_wtext;
	r.height = ib->ib_htext;

	drawtext(vid,ib->ib_ptext,&r,2,ib->ib_char >> 12,5,MD_REPLACE,0);

	r.x = x + ib->ib_xchar;
	r.y = y + ib->ib_ychar;
	r.width = 0;
	r.height = 0;
	
	ch[0] = ib->ib_char & 0x00ff;
	ch[1] = 0;

	drawtext(vid,ch,&r,0,ib->ib_char >> 12,5,MD_TRANS,0);
}

static void drawcicon(WORD vid,WORD x,WORD y,CICONBLK *cib) {
	char	ch[2];
	
	MFDB	d,s;

	WORD	xyarray[8];
	WORD	color[] = {WHITE,BLACK};
	
	RECT	r;
	CICON *best = NULL,*ciwalk = cib->mainlist;
	WORD  bestplanes = 0;

	while(ciwalk) {
		if((ciwalk->num_planes <= num_planes) &&
			(ciwalk->num_planes > bestplanes)) {
			best = ciwalk;
			bestplanes = best->num_planes;
			
			if(bestplanes == num_planes) {
				break;
			};
		};
		
		ciwalk = ciwalk->next_res;
	};

	if(!best) {
		drawicon(vid,x,y,&cib->monoblk);
	}
	else {
		xyarray[0] = 0;
		xyarray[1] = 0;
		xyarray[2] = cib->monoblk.ib_wicon - 1;
		xyarray[3] = cib->monoblk.ib_hicon - 1;
		xyarray[4] = x + cib->monoblk.ib_xicon;
		xyarray[5] = y + cib->monoblk.ib_yicon;
		xyarray[6] = x + cib->monoblk.ib_xicon + cib->monoblk.ib_wicon - 1;
		xyarray[7] = y + cib->monoblk.ib_yicon + cib->monoblk.ib_hicon - 1;
	
		(LONG)s.fd_addr = (LONG)best->col_mask;
		s.fd_w = cib->monoblk.ib_wicon;
		s.fd_h = cib->monoblk.ib_hicon;
		s.fd_wdwidth = ((cib->monoblk.ib_wicon +15) >> 4);
		s.fd_stand = 0;
		s.fd_nplanes = 1;
	
		(LONG)d.fd_addr = 0L;
	
		Vdi_vrt_cpyfm(vid,MD_TRANS,xyarray,&s,&d,color);

		s.fd_nplanes = best->num_planes;	
		(LONG)s.fd_addr = (LONG)best->col_data;
		
		Vdi_vro_cpyfm(vid,S_OR_D,xyarray,&s,&d);
	
		r.x = x + cib->monoblk.ib_xtext;
		r.y = y + cib->monoblk.ib_ytext;
		r.width = cib->monoblk.ib_wtext;
		r.height = cib->monoblk.ib_htext;
	
		drawtext(vid,cib->monoblk.ib_ptext,&r,2,cib->monoblk.ib_char >> 12,5,MD_REPLACE,0);
	
		r.x = x + cib->monoblk.ib_xchar;
		r.y = y + cib->monoblk.ib_ychar;
		r.width = 0;
		r.height = 0;
		
		ch[0] = cib->monoblk.ib_char & 0x00ff;
		ch[1] = 0;
	
		drawtext(vid,ch,&r,0,cib->monoblk.ib_char >> 12,5,MD_TRANS,0);
	};
}

static void	drawrsc(WORD vid,OBJECT *tree,WORD first, WORD drawbrsi, WORD depth
	,RECT *clip,WORD x, WORD y) {
	WORD	last;
	
	do {
		if(!(tree[first].ob_flags & HIDETREE)) {
			char	text[2];
			RECT	r;
			TIEXT	*dtiext;
			WORD  type = tree[first].ob_type & 0xff;
					
			switch(type) {
				case G_BOX		:	/*0x14*/
					{
						Vdi_vsl_color(vid,tree[first].ob_spec.obspec.framecol);

						r.x = tree[first].ob_x + x;
						r.y = tree[first].ob_y + y;
						r.width = tree[first].ob_width;
						r.height = tree[first].ob_height;
						
						setwritemode(vid,SWRM_REPLACE);
						Vdi_vsl_type(vid,1);
		
						if(tree[first].ob_state & SELECTED) {
							Vdi_vsf_color(vid,BLACK);
							setfilltype(vid,7);
					
							drawrectfi(vid,&r,0);
								
							setwritemode(vid,SWRM_TRANS);
		
							Vdi_vsf_color(vid,invertcolor(tree[first].ob_spec.obspec.interiorcol));
						}
						else {
							Vdi_vsf_color(vid,tree[first].ob_spec.obspec.interiorcol);
						};
							
						setfilltype(vid,tree[first].ob_spec.obspec.fillpattern);
						
						if(tree[first].ob_state & OUTLINED) {
							RECT r2;
	
							Objc_area_needed(tree,first,&r2);
							
							drawrectfi(vid,&r2,1);
							
							drawframe(vid,&r,tree[first].ob_spec.obspec.framesize);
						}
						else {
							drawrectfi(vid,&r,tree[first].ob_spec.obspec.framesize);
						};
					};
					break;
					
				case G_TEXT:    /*0x15*/
				case G_BOXTEXT: /*0x16*/

					dtiext = (TIEXT *)&tree[first].ob_spec.tedinfo->te_color;

					r.x = tree[first].ob_x + x;
					r.y = tree[first].ob_y + y;
					r.width = tree[first].ob_width;
					r.height = tree[first].ob_height;

					if(type == G_BOXTEXT) {
						Vdi_vsl_color(vid,dtiext->framecol);
						
						setwritemode(vid,SWRM_REPLACE);
						Vdi_vsl_type(vid,1);
	
						if(tree[first].ob_state & SELECTED) {
							Vdi_vsf_color(vid,BLACK);
							setfilltype(vid,7);
				
							drawrectfi(vid,&r,tree[first].ob_spec.tedinfo->te_thickness);
							
							setwritemode(vid,SWRM_TRANS);
	
							Vdi_vsf_color(vid,invertcolor(dtiext->interiorcol));
						}
						else {
							Vdi_vsf_color(vid,dtiext->interiorcol);
						};
						
						setfilltype(vid,dtiext->fillpattern);
						drawrectfi(vid,&r,tree[first].ob_spec.tedinfo->te_thickness);
					};
					
					drawtext(vid,tree[first].ob_spec.tedinfo->te_ptext
						,&r,tree[first].ob_spec.tedinfo->te_just
						,dtiext->textcol
						,tree[first].ob_spec.tedinfo->te_font
						,dtiext->textmode
						,tree[first].ob_state);
					break;
					
				case G_IMAGE: /*0x17*/
					drawimage(vid,tree[first].ob_x + x,tree[first].ob_y + y
						,(BITBLK *)tree[first].ob_spec.index);
					break;
			
				case G_PROGDEF: /*0x18*/
					{
						PARMBLK pb;
						WORD    remainstate;
						RECT    oclip;
						
						pb.pb_tree = tree;
						pb.pb_obj = first;
						pb.pb_prevstate = tree[first].ob_state;
						pb.pb_currstate = tree[first].ob_state;
						
						Objc_do_offset(tree,first,(WORD *)&pb.pb_x);
						pb.pb_w = tree[first].ob_width;
						pb.pb_h = tree[first].ob_height;
						
						Objc_calc_clip(tree,first,&oclip);
						
						if(Misc_intersect(&oclip,clip,(RECT *)&pb.pb_xc)) {
							pb.pb_parm = ((USERBLK *)tree[first].ob_spec.index)->ub_parm;
					
							remainstate = ((USERBLK *)tree[first].ob_spec.index)->ub_code(&pb);
						};
					};
					break;

				case G_IBOX	:	/*0x19*/
					{
						Vdi_vswr_mode(vid,MD_REPLACE);
						Vdi_vsl_color(vid,tree[first].ob_spec.obspec.framecol);
						r.x = tree[first].ob_x + x;
						r.y = tree[first].ob_y + y;
						r.width = tree[first].ob_width;
						r.height = tree[first].ob_height;
						drawframe(vid,&r,tree[first].ob_spec.obspec.framesize);
					};
					break;
					
				case G_BUTTON	:	/*0x1a*/
					setwritemode(vid,1);
					if(tree[first].ob_state & SELECTED) {
						Vdi_vsf_color(vid,BLACK);
					}
					else {
						Vdi_vsf_color(vid,WHITE);
					};
					
					setfilltype(vid,7);
					Vdi_vsl_color(vid,BLACK);
					r.x = tree[first].ob_x + x;
					r.y = tree[first].ob_y + y;
					r.width = tree[first].ob_width;
					r.height = tree[first].ob_height;
					
					if(tree[first].ob_flags & DEFAULT) {
						drawrectfi(vid,&r,-DEFBUTFRAME);
					}
					else {
						drawrectfi(vid,&r,-BUTTONFRAME);
					};
									
					drawtext(vid,(char *)tree[first].ob_spec.index,&r,2,BLACK
						,3,0,tree[first].ob_state);
					break;
			
				case G_BOXCHAR	:	/*0x1b*/
					r.x = tree[first].ob_x + x;
					r.y = tree[first].ob_y + y;
					r.width = tree[first].ob_width;
					r.height = tree[first].ob_height;

					Vdi_vsf_color(vid,tree[first].ob_spec.obspec.framecol);

					setwritemode(vid,SWRM_REPLACE);
					Vdi_vsl_type(vid,1);

					if(tree[first].ob_state & SELECTED) {
						Vdi_vsf_color(vid,BLACK);
						setfilltype(vid,7);
			
						drawrectfi(vid,&r,0);

						setwritemode(vid,SWRM_TRANS);

						Vdi_vsf_color(vid,invertcolor(tree[first].ob_spec.obspec.interiorcol));
					}
					else {
						Vdi_vsf_color(vid,tree[first].ob_spec.obspec.interiorcol);
					};
					
					setfilltype(vid,tree[first].ob_spec.obspec.fillpattern);
					drawrectfi(vid,&r,tree[first].ob_spec.obspec.framesize);
					
					text[0] = tree[first].ob_spec.obspec.character;
					text[1] = 0;
					drawtext(vid,text,&r,2,tree[first].ob_spec.obspec.textcol
						,3,tree[first].ob_spec.obspec.textmode
						,tree[first].ob_state);
					break;
					
				case G_STRING	:	/*0x001c*/
				case G_TITLE	:	/*0x0020*/
					{
						r.x = tree[first].ob_x + x;
						r.y = tree[first].ob_y + y;
						r.width = tree[first].ob_width;
						r.height = tree[first].ob_height;
	
						drawtext(vid,tree[first].ob_spec.free_string
							,&r,0
							,BLACK
							,3
							,1
							,tree[first].ob_state);					
					};
					
					break;
					
				case G_FTEXT:    /*0x1d*/
				case G_FBOXTEXT: /*0x1e*/
					{
						char ctext[100];
						char *ptmplt = tree[first].ob_spec.tedinfo->te_ptmplt;
						char *ptext = tree[first].ob_spec.tedinfo->te_ptext;
						char *pctext = ctext;
						
						while(*ptmplt) {
							switch(*ptmplt) {
							case '_':
								if(*ptext) {
									*(pctext++) = *(ptext++);
								}
								else {
									*(pctext++) = '_';
								};

								ptmplt++;
								break;
								
							default:
								*(pctext++) = *(ptmplt++);
							};
						};
						
						*pctext = '\0';
						
						dtiext = (TIEXT *)&tree[first].ob_spec.tedinfo->te_color;
	
						r.x = tree[first].ob_x + x;
						r.y = tree[first].ob_y + y;
						r.width = tree[first].ob_width;
						r.height = tree[first].ob_height;
	
						if(type == G_FBOXTEXT) {
							Vdi_vsl_color(vid,dtiext->framecol);
							
							setwritemode(vid,SWRM_REPLACE);
							Vdi_vsl_type(vid,1);
		
							if(tree[first].ob_state & SELECTED) {
								Vdi_vsf_color(vid,BLACK);
								setfilltype(vid,7);
					
								drawrectfi(vid,&r,tree[first].ob_spec.tedinfo->te_thickness);
								
								setwritemode(vid,SWRM_TRANS);
		
								Vdi_vsf_color(vid,invertcolor(dtiext->interiorcol));
							}
							else {
								Vdi_vsf_color(vid,dtiext->interiorcol);
							};
							
							setfilltype(vid,dtiext->fillpattern);
							drawrectfi(vid,&r,tree[first].ob_spec.tedinfo->te_thickness);
						};
						
						drawtext(vid,ctext
							,&r,tree[first].ob_spec.tedinfo->te_just
							,dtiext->textcol
							,tree[first].ob_spec.tedinfo->te_font
							,dtiext->textmode
							,tree[first].ob_state);
					};
					break;
					
				case	G_ICON:
					drawicon(vid,tree[first].ob_x + x,tree[first].ob_y + y
						,tree[first].ob_spec.iconblk);
					break;
				
				case	G_CICON:
					drawcicon(vid,tree[first].ob_x + x,tree[first].ob_y + y
						,(CICONBLK *)tree[first].ob_spec.index);
					
					break;
				
				default:
					fprintf(stderr,"%s: Line %d: drawrsc:\r\n"
					"Not implemented object type (%d)!\r\n"
						,__FILE__,__LINE__,
						tree[first].ob_type);
			}

			if(tree[first].ob_state & CHECKED) {
				WORD dum;
				
				Vdi_vst_alignment(vid,0,5,&dum,&dum);
				Vdi_v_gtext(vid,tree[first].ob_x + x,tree[first].ob_y + y,
					"\x8");
			};
			
			if((depth) && (tree[first].ob_head > 0)) {
				drawrsc(vid,tree,tree[first].ob_head,TRUE,depth - 1,clip
					,x + tree[first].ob_x,y + tree[first].ob_y);
			};
		};
		
		if(!drawbrsi)
			break;
			
		last = first;
		if((first = tree[first].ob_next) < 0)
			break;
	}while(tree[first].ob_tail != last);
}

static void get_char_bound(OBJECT *tree,WORD obj,WORD first,WORD last,
														RECT *r) {
	char *ptmplt = tree[obj].ob_spec.tedinfo->te_ptmplt;
	WORD firstreal = 0,lastreal = 0;
	
	while(*ptmplt) {
		if(*ptmplt == '_') {
			first--;
			last--;
		};

		if(first >= 0) {
			firstreal++;
		};
		
		if(last >= 0) {
			lastreal++;
		};
		
		ptmplt++;
	};

	Objc_do_offset(tree,obj,(WORD *)r);

	r->y += (tree[obj].ob_height - largeth) >> 1;
	r->height = largeth;
	r->width = largetw * (lastreal - firstreal);
	
	switch(tree[obj].ob_spec.tedinfo->te_just) {
	case	0	:
		break;
	case	1	:
		r->x += (tree[obj].ob_width - 
						(largetw * (WORD)strlen(tree[obj].ob_spec.tedinfo->te_ptmplt)));
		break;
	case	2	:
		r->x += ((tree[obj].ob_width -
						(largetw * (WORD)strlen(tree[obj].ob_spec.tedinfo->te_ptmplt))) >> 1);
		break;
	};

	r->x = r->x + (firstreal * largetw);
}

static void draw_cursor(WORD vid,WORD pos,OBJECT *tree,WORD obj) {
	RECT r;
	WORD xy[4];
	
	get_char_bound(tree,obj,pos,pos,&r);

	xy[0] = r.x;
	xy[1] = r.y;
	xy[2] = xy[0];
	xy[3] = xy[1] + r.height;

	Vdi_vswr_mode(vid,MD_XOR);
	Vdi_v_pline(vid,2,xy);
}

static WORD handle_ed_char(WORD vid,WORD idx,OBJECT *tree,WORD obj,
						WORD kc) {
	RECT clip;
	
	switch(kc) {
	case 0x0e08: /* backspace */
	  if(idx > 0) {
	    draw_cursor(vid,idx,tree,obj);
			
	    strcpy(&tree[obj].ob_spec.tedinfo->te_ptext[idx - 1],
		   &tree[obj].ob_spec.tedinfo->te_ptext[idx]);
	    
	    get_char_bound(tree,obj,idx - 1,
			   (WORD)strlen(tree[obj].ob_spec.tedinfo->te_ptext)
			   + 2,&clip);

	    Objc_do_draw(vid,tree,0,9,&clip);
			
	    draw_cursor(vid,idx - 1,tree,obj);
	    return idx - 1;
	  };
	  return idx;
		
	case 0x4b00: /* left */
		if(idx > 0) {
			draw_cursor(vid,idx,tree,obj);
			draw_cursor(vid,idx - 1,tree,obj);
			return idx - 1;
		};
		return idx;
	
	case 0x4d00: /* right */
		if(idx < strlen(tree[obj].ob_spec.tedinfo->te_ptext)) {
			draw_cursor(vid,idx,tree,obj);
			draw_cursor(vid,idx + 1,tree,obj);
			return idx + 1;
		};
		return idx;

	case 0x537f: /* delete */
		if(idx < strlen(tree[obj].ob_spec.tedinfo->te_ptext)) {
			draw_cursor(vid,idx,tree,obj);
			
			strcpy(&tree[obj].ob_spec.tedinfo->te_ptext[idx],
							&tree[obj].ob_spec.tedinfo->te_ptext[idx + 1]);

			get_char_bound(tree,obj,idx,
				(WORD)strlen(tree[obj].ob_spec.tedinfo->te_ptext) + 1,&clip);
			Objc_do_draw(vid,tree,0,9,&clip);
			
			draw_cursor(vid,idx,tree,obj);
		};
		return idx;
		
	default:
		if(tree[obj].ob_spec.tedinfo->te_ptext[idx] != '\0') {
			char i = strlen(tree[obj].ob_spec.tedinfo->te_ptext);
			
			while(i >= idx) {
				tree[obj].ob_spec.tedinfo->te_ptext[i + 1] = 
					tree[obj].ob_spec.tedinfo->te_ptext[i];
				
				i--;
			};
		}
		else {
			tree[obj].ob_spec.tedinfo->te_ptext[idx + 1] = '\0';
		};
		
		tree[obj].ob_spec.tedinfo->te_ptext[idx] = (char)kc;

		draw_cursor(vid,idx,tree,obj);
		get_char_bound(tree,obj,idx,
										(WORD)strlen(tree[obj].ob_spec.tedinfo->te_ptext),&clip);
		Objc_do_draw(vid,tree,0,9,&clip);
		draw_cursor(vid,idx + 1,tree,obj);
		
		return idx + 1;
	};
}

/****************************************************************************
 * Public functions                                                         *
 ****************************************************************************/

void	init_objc(void) {	
	WORD work_out[57];
	
	Vdi_vq_extnd(globals.vid,1,work_out);
	
	num_planes = work_out[4];
	
	maxnumcols = 16;	/* <-- NDRA!!!*/
		
	Vdi_v_show_c(globals.vid,0);

	largeth = globals.cheight;
	largetw = globals.cwidth;
	smallth = globals.cheight >> 1;
	textyoffs = globals.cheight >> 1;
}

void	exit_objc(void) {
}

void	do_objc_add(OBJECT *t,WORD p,WORD c) {
	if(t[p].ob_tail < 0) {
		t[p].ob_head = c;
		t[p].ob_tail = c;
		t[c].ob_next = p;
	}
	else
	{
		t[c].ob_next = p;
		t[t[p].ob_tail].ob_next = c;
		t[p].ob_tail = c;
	};
}

/*objc_add	0x0028*/

void	Objc_add(AES_PB *apb)
{
	do_objc_add((OBJECT *)apb->addr_in[0],apb->int_in[0]
		,apb->int_in[1]);
	
	apb->int_out[0] = 1;
}

/*objc_draw	0x002a*/

/****************************************************************************
 * Objc_do_draw                                                             *
 *  Implementation of objc_draw().                                          *
 ****************************************************************************/
WORD              /* 0 if an error occured, or 1.                           */
Objc_do_draw(     /*                                                        */
WORD   vid,       /* VDI workstation id.                                    */
OBJECT *t,        /* Resource tree.                                         */
WORD   start,     /* Start object.                                          */
WORD   depth,     /* Maximum draw depth.                                    */
RECT   *xywh)     /* Clipping rectangle.                                    */
/****************************************************************************/
{
	WORD	xy[2];
	WORD	xyxy[4];

	if(Objc_do_offset(t,start,xy) == 0) {
		return 0;
	};
	
	xy[0] -= t[start].ob_x;
	xy[1] -= t[start].ob_y;

	xyxy[0] = xywh->x;
	xyxy[1] = xywh->y;
	xyxy[2] = xyxy[0] + xywh->width - 1;
	xyxy[3] = xyxy[1] + xywh->height - 1;
	
	Vdi_vs_clip(vid,1,xyxy);

	Vdi_v_hide_c(vid);
	
	drawrsc(vid,t,start,0
		,depth,xywh,xy[0],xy[1]);

	Vdi_v_show_c(vid,1);

	Vdi_vs_clip(vid,0,xyxy);
	
	return 1;
}

void	Objc_draw(AES_PB *apb) {
	WORD	vid = Appl_get_vid(apb->global[2]);
	
	if(vid > 0) {
		apb->int_out[0] = Objc_do_draw(vid,(OBJECT *)apb->addr_in[0],apb->int_in[0]
			,apb->int_in[1],(RECT *)&apb->int_in[2]);
	}
	else {
		apb->int_out[0] = 0;
		fprintf(stderr,"%s: Line %d: Couldn't find application description!\r\n",
			__FILE__,__LINE__);
	};
}

/*objc_find 0x002b*/

/****************************************************************************
 * Objc_do_find                                                             *
 *  Implementation of objc_find().                                          *
 ****************************************************************************/
WORD              /* Object index, or -1.                                   */
Objc_do_find(     /*                                                        */
OBJECT *t,        /* Resource tree to search.                               */
WORD startobject, /* Start object.                                          */
WORD depth,       /* Maximum depth.                                         */
WORD x,           /* X offset.                                              */
WORD y,           /* Y offset.                                              */
WORD level)       /* Current depth of search.                               */
/****************************************************************************/
{
	if(t[startobject].ob_flags & HIDETREE) {
		return -1;
	};

	if(!level) {
		WORD	lxy[2];
		Objc_do_offset(t,startobject,lxy);

		x -= lxy[0];
		y -= lxy[1];
	};
	
	if((x >= 0) && (x < t[startobject].ob_width) 
		&& (y >= 0) && (y < t[startobject].ob_height)) {
		WORD	deeper;
			
		if((depth > 0) && (t[startobject].ob_head >= 0)) {
			WORD i = t[startobject].ob_head;
			
			while(i != startobject) {
				deeper = Objc_do_find(t,i,depth - 1,
					x - t[i].ob_x,
					y - t[i].ob_y,level + 1);
	
				if(deeper >= 0)
					return deeper;

				i = t[i].ob_next;
			};
		};
		
		return startobject;
	};
		
	return -1;
}

void	Objc_find(AES_PB *apb) {
	apb->int_out[0] = Objc_do_find((OBJECT *)apb->addr_in[0],apb->int_in[0],
		apb->int_in[1],apb->int_in[2],apb->int_in[3],0);
};

/*objc_offset 0x002c*/

/****************************************************************************
 * Objc_do_offset                                                           *
 *  Implementation of objc_offset().                                        *
 ****************************************************************************/
WORD              /* 0 if error, or 1.                                      */
Objc_do_offset(   /*                                                        */
OBJECT *t,        /* Resource tree.                                         */
WORD   o,         /* Object index.                                          */
WORD   *xy)       /* X and Y coordinates of object if successfull.          */
/****************************************************************************/
{
	if(!t) {
		return 0;
	};
	
	xy[0] = 0;
	xy[1] = 0;
	
	do {
		WORD	last;
		
		xy[0] += t[o].ob_x;
		xy[1] += t[o].ob_y;
		
		if(t[o].ob_next < 0) {
			break;
		};
		
		do {
			last = o;
			o = t[o].ob_next;
		}while(last != t[o].ob_tail);	
	}while(1);
	
	if(!o) {
		return 1;
	}
	else {
		return 0;
	};
}


void	Objc_offset(AES_PB *apb) {
	apb->int_out[0] = Objc_do_offset((OBJECT *)apb->addr_in[0]
		,apb->int_in[0], &apb->int_out[1]);
};

/****************************************************************************
 * Objc_do_edit                                                             *
 *  Implementation of objc_edit().                                          *
 ****************************************************************************/
WORD              /* 0 if an error occured, or 1.                           */
Objc_do_edit(     /*                                                        */
WORD   vid,       /* VDI workstation handle.                                */
OBJECT *tree,     /* Resource tree.                                         */
WORD   obj,       /* Object index.                                          */
WORD   kc,        /* Key code to process.                                   */
WORD   *idx,      /* Character index.                                       */
WORD   mode)      /* Edit mode.                                             */
/****************************************************************************/
{
	switch(mode) {
	case ED_INIT:
		*idx = (WORD)strlen(tree[obj].ob_spec.tedinfo->te_ptext);

		draw_cursor(vid,*idx,tree,obj);

		return 1;
	
	case ED_CHAR:
		*idx = handle_ed_char(vid,*idx,tree,obj,kc);
		return 1;
		
	case ED_END:
		draw_cursor(vid,*idx,tree,obj);
		
		return 1;
		
	default:
		fprintf(stderr,"%s: Line %d: Objc_do_edit:\r\n"
						"Unknown mode %d\r\n",__FILE__,__LINE__,mode);
		return 0;
	}
}

/****************************************************************************
 * Objc_edit                                                                *
 *   0x002e objc_edit().                                                    *
 ****************************************************************************/
void              /*                                                        */
Objc_edit(        /*                                                        */
AES_PB *apb)      /* AES parameter block.                                   */
/****************************************************************************/
{
	WORD vid = Appl_get_vid(apb->global[2]);
	
	apb->int_out[1] = apb->int_in[2];
	apb->int_out[0] = Objc_do_edit(vid,
										(OBJECT *)apb->addr_in[0],apb->int_in[0],
										apb->int_in[1],&(apb->int_out[1]),
										apb->int_in[3]);
}

/****************************************************************************
 * Objc_do_change                                                           *
 *  Implementation of objc_change().                                        *
 ****************************************************************************/
WORD              /* 0 if an error occured, or 1.                           */
Objc_do_change(   /*                                                        */
WORD   vid,       /* VDI workstation handle.                                */
OBJECT *tree,     /* Resource tree.                                         */
WORD   obj,       /* Object index.                                          */
RECT   *clip,     /* Clipping rectangle.                                    */
WORD   newstate,  /* New object state.                                      */
WORD   drawflag)  /* Drawing flag.                                          */
/****************************************************************************/
{
	if(vid == -1) {
		return 0;
	};
	
	tree[obj].ob_state = newstate;
	
	if(drawflag == REDRAW) {
		Objc_do_draw(vid,tree,obj,9,clip);
	};
	
	return 1;
}

/****************************************************************************
 * Objc_change                                                              *
 *   0x002f objc_change().                                                  *
 ****************************************************************************/
void              /*                                                        */
Objc_change(      /*                                                        */
AES_PB *apb)      /* AES parameter block.                                   */
/****************************************************************************/
{
	WORD vid = Appl_get_vid(apb->global[2]);
	
	apb->int_out[0] = Objc_do_change(vid,
										(OBJECT *)apb->addr_in[0],apb->int_in[0],
										(RECT *)&apb->int_in[2],
										apb->int_in[6],apb->int_in[7]);
}

/****************************************************************************
 *  Objc_area_needed                                                        *
 *   Calculate how large area an object covers.                             *
 ****************************************************************************/
void              /*                                                        */
Objc_area_needed( /*                                                        */
OBJECT *tree,     /* Pointer to the root of the resource tree.              */
WORD   object,    /* Index of interesting object.                           */
RECT   *rect)     /* Buffer where the requested area size will be placed.   */
/****************************************************************************/
{
	Objc_do_offset(tree,object,(WORD *)rect);
	
	rect->width = tree[object].ob_width;
	rect->height = tree[object].ob_height;
	
	switch(tree[object].ob_type) {
		case	G_BOX:
		case	G_IBOX:
		case	G_BOXCHAR:
			if(tree[object].ob_spec.obspec.framesize < 0) {
				rect->x += tree[object].ob_spec.obspec.framesize;
				rect->y += tree[object].ob_spec.obspec.framesize;
				rect->width -= (tree[object].ob_spec.obspec.framesize << 1);
				rect->height -= (tree[object].ob_spec.obspec.framesize << 1);
			};
			break;
		case	G_BUTTON:
			if(tree[object].ob_flags & DEFAULT) {
				rect->x -= DEFBUTFRAME;
				rect->y -= DEFBUTFRAME;
				rect->width += (DEFBUTFRAME << 1);
				rect->height += (DEFBUTFRAME << 1);
			}
			else {
				rect->x -= BUTTONFRAME;
				rect->y -= BUTTONFRAME;
				rect->width += (BUTTONFRAME << 1);
				rect->height += (BUTTONFRAME << 1);
			};
			break;
	};
	
	if(tree[object].ob_state & OUTLINED) {
		rect->x -= OUTLINESIZE;
		rect->y -= OUTLINESIZE;
		rect->width += (OUTLINESIZE << 1);
		rect->height += (OUTLINESIZE << 1);
	};
}

/****************************************************************************
 *  Objc_calc_clip                                                          *
 *   Calculate required clip area for object.                               *
 ****************************************************************************/
void              /*                                                        */
Objc_calc_clip(   /*                                                        */
OBJECT *tree,     /* Pointer to the root of the resource tree.              */
WORD   object,    /* Index of interesting object.                           */
RECT   *rect)     /* Buffer where the requested area size will be placed.   */
/****************************************************************************/
{
	WORD owalk = object;
	
	while(1) {
		if(tree[owalk].ob_next == -1) {
			owalk = -1;
			break;
		};
		
		if(tree[tree[owalk].ob_next].ob_tail == owalk) {
			owalk = tree[owalk].ob_next;
			break;
		};
		
		owalk = tree[owalk].ob_next;
	};
	
	if(owalk == -1) {
		Objc_area_needed(tree,object,rect);
	}
	else {
		Objc_do_offset(tree,owalk,(WORD *)rect);
	
		rect->width = tree[owalk].ob_width;
		rect->height = tree[owalk].ob_height;
	};
}
